/*
 * The contents of this file are subject to the terms
 * of the Common Development and Distribution License
 * (the License).  You may not use this file except in
 * compliance with the License.
 *
 * You can obtain a copy of the license at
 * https://glassfish.dev.java.net/public/CDDLv1.0.html or
 * glassfish/bootstrap/legal/CDDLv1.0.txt.
 * See the License for the specific language governing
 * permissions and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL
 * Header Notice in each file and include the License file
 * at glassfish/bootstrap/legal/CDDLv1.0.txt.
 * If applicable, add the following below the CDDL Header,
 * with the fields enclosed by brackets [] replaced by
 * you own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 */

package com.sun.tools.xjc.addon.property_listener_injector;

import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JFieldVar;
import com.sun.codemodel.JInvocation;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.tools.xjc.BadCommandLineException;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.Plugin;
import com.sun.tools.xjc.model.CPluginCustomization;
import com.sun.tools.xjc.outline.ClassOutline;
import com.sun.tools.xjc.outline.Outline;
import com.sun.tools.xjc.util.DOMUtils;
import org.xml.sax.ErrorHandler;

import javax.xml.bind.annotation.XmlTransient;
import java.beans.VetoableChangeListener;
import java.beans.VetoableChangeSupport;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Collections;
import java.util.List;

/**
 * This Plugin will generate property change events on each setXXX.
 *
 * See the javadoc of {@link Plugin} for what those methods mean.
 *
 * @author Jerome Dochez
 */
public class PluginImpl extends Plugin {
    public String getOptionName() {
        return "Xinject-listener-code";
    }

    public List<String> getCustomizationURIs() {
        return Collections.singletonList(Const.NS);
    }

    public boolean isCustomizationTagName(String nsUri, String localName) {
        return nsUri.equals(Const.NS) && localName.equals("listener");
    }

    public String getUsage() {
        return "  -Xinject-listener-code\t:  inject property change event support to setter methods";
    }

    public void onActivated(Options opt) throws BadCommandLineException {
        FieldRendererFactory frf = new FieldRendererFactory(opt.getFieldRendererFactory());
        opt.setFieldRendererFactory(frf, this);
    }
    
    // meat of the processing
    public boolean run(Outline model, Options opt, ErrorHandler errorHandler) {
        
        
        for( ClassOutline co : model.getClasses() ) {
            CPluginCustomization c = co.target.getCustomizations().find(Const.NS,"listener");
            //if(c==null)
            //    continue;   // no customization --- nothing to inject here

            //c.markAsAcknowledged();
            // TODO: ideally you should validate this DOM element to make sure
            // that there's no typo/etc. JAXP 1.3 can do this very easily.
            String interfaceName;
            if (c!=null) {
                interfaceName = DOMUtils.getElementText(c.element);
            } else {
                interfaceName = VetoableChangeListener.class.getName();
            }
            if (VetoableChangeListener.class.getName().equals(interfaceName)) {
                addSupport(VetoableChangeListener.class, VetoableChangeSupport.class, co.implClass);
            }
            
        }

        return true;
    }
    
    private void addSupport(Class listener, Class support, JDefinedClass target) {
        
        // add the support field.
//        JFieldVar field = target.field(JMod.PRIVATE|JMod.TRANSIENT, support, "support");
        JFieldVar field = target.field(JMod.PRIVATE, support, "support");
        field.annotate(XmlTransient.class);
        
        // and initialize it....
        field.init(JExpr.direct("new " + support.getSimpleName() + "(this)"));
        
        // we need to hadd
        for (Method method : support.getMethods()) {
            if (Modifier.isPublic(method.getModifiers())) {
                if (method.getName().startsWith("add")) {
                    // look if one of the parameters in the listnener class...
                    for (Class param : method.getParameterTypes()) {
                        if (param.getName().equals(listener.getName())) {
                            addMethod(method, target);
                            break;
                        }
                    }
                }
                if (method.getName().startsWith("remove")) {
                    // look if one of the parameters in the listnener class...
                    for (Class param : method.getParameterTypes()) {
                        if (param.getName().equals(listener.getName())) {
                            addMethod(method, target);
                            break;
                        }
                    }
                }                
            }
        }
    }
    
    private void addMethod(Method method, JDefinedClass target) {
        JMethod addListener = target.method(JMod.PUBLIC, method.getReturnType(), method.getName());
        Class params[] = method.getParameterTypes();

        JInvocation inv = addListener.body().invoke(JExpr.ref("support"), method.getName());

        for (int i=0;i<params.length;i++) {
            inv.arg(addListener.param(params[i], "param"+i));
        }
    }
}
